#include <opencv2/opencv.hpp>
#include <opencv2/core/ocl.hpp>
#include <cstdlib>
#include <cstdio>
#include <math.h>
#include <iostream>
#include <condition_variable>
#include <mutex>
#include <thread>
#include <vector>
#include "time.h"

#define NUM_FRAMES 200
#ifdef DEBUG
#define DEBUG(X) X
#else
#define DEBUG(X)
#endif

#define OPENCL false

using namespace std;
using namespace cv;

string window_name = "Edge Map";

// Circular Buffer class
template <class ElementType>
class CircularBuffer {
private:
    int numElements;
    int bufferSize;
    int Head;
    int Tail;
    mutex mtx;
    condition_variable cv;
    vector<ElementType*> buffer;
    string name;

public:
    CircularBuffer(int maxElements, string _name = "ANON_BUFFER") {
        Head = 0;
        Tail = 0;
        numElements = 0;
        bufferSize = maxElements+1;
        name = _name;
        buffer.reserve(bufferSize);
    }

    bool push(ElementType* E, string caller);
    ElementType* pop(string caller);

};

template <class ElementType>
bool CircularBuffer<ElementType>::push(ElementType* E, string caller) {
    unique_lock<mutex> lk(mtx);
    if((Head +1) % bufferSize == Tail) {
        cv.wait(lk);
    }
    buffer[Head] = E;
    Head = (Head+1) % bufferSize;
    numElements++;
    DEBUG(cout << "[" << caller << "]: " << name << " Total Elements = " << numElements << "\n");
    lk.unlock();
    cv.notify_one();
    return true;
}

template <class ElementType>
ElementType* CircularBuffer<ElementType>::pop(string caller) {
    unique_lock<mutex> lk(mtx);
    if(Tail == Head) {
        //DEBUG(cout << "[C] Going to sleep ...\n");
        cv.wait(lk);
        //DEBUG(cout << "[C] Waking up\n");
    }
    ElementType* E = buffer[Tail];
    Tail = (Tail + 1) % bufferSize;
    numElements--;
    DEBUG(cout << "[" << caller << "]: " << name << " Total Elements = " << numElements << "\n");
    lk.unlock();
    cv.notify_one();
    return E;
}
// --------------------------------------------------------------------------


// Read image from a file and convert it into gray sclae
Mat readImage(char* file) {
    Mat src = imread( file );

    if( !src.data )
    {
        DEBUG(cout << "Error: Canot read input image " << file << "\n");
        exit(-1);
    }

    /// Convert the image to grayscale
    Mat src_gray;
    cvtColor( src, src, CV_BGR2GRAY );
    src.convertTo(src, CV_32F, 1.0/255.0);
    return src;
}

// Gaussian Smoothening Node function
void GaussianSmoothening(float Sn, CircularBuffer<Mat>* in_I, CircularBuffer<Mat>* out_IBlur, float* time) {
    Timestamp start, end;
    start = get_time(TIMER_THREAD);
    ocl::setUseOpenCL(OPENCL);
    DEBUG(cout << "Gaussian Smoothening Starts\n");
    string id = "GS";
    int i = 0;
    while(i < NUM_FRAMES) {
        //waitKey(0);
        Mat* I = in_I->pop(id);
        DEBUG(cout << "[" << id << "]: Iteration " << i << "\n");
        int kernelSize = 2*ceil(3*Sn) + 1;
        Mat* IBlur = new Mat(I->rows, I->cols, I->type());
        GaussianBlur(*I, *IBlur, Size(kernelSize, kernelSize), Sn);
        out_IBlur->push(IBlur, id);
        I->release();
        i++;
    }
    end = get_time(TIMER_THREAD);
    *time = (end-start)*1.0/BILLION;
    DEBUG(cout << "Gaussian Smoothening Done\n");
}

// Laplacian Node
void NonLinearLaplacian(Mat B, CircularBuffer<Mat>* in_IBlur, CircularBuffer<Mat>* out_L, float* time, float* ctime) {
    Timestamp start, end, t1, t2, get=0, copy =0, ed=0, allocation=0, calc=0, copyB=0, sub=0, push=0, release=0;

    start = get_time(TIMER_THREAD);
    ocl::setUseOpenCL(true);
    DEBUG(cout << "Non-linear Laplacian Starts\n");
    string id = "L";
    int i = 0;
    while(i < NUM_FRAMES) {
        //waitKey(0);

        // Waiting for data
        //t1 = get_time(TIMER_MT);
        Mat* IBlur = in_IBlur->pop(id);
        //t2 = get_time(TIMER_MT);
        //get += t2-t1;
        DEBUG(cout << "[" << id << "]: Iteration " << i << "\n");

        // Copy to GPU
        UMat IErode, IDilate;
        //t1 = get_time(TIMER_MT);
        UMat IBlurU = IBlur->getUMat(ACCESS_READ);
        //t2 = get_time(TIMER_MT);
        //copy += t2-t1;
        
        // Erode dilate calculation
        // Non-Linear Laplacian Estimate
        //t1 = get_time(TIMER_MT);
        erode(IBlurU, IErode, B);
        dilate(IBlurU, IDilate, B);
        //t2 = get_time(TIMER_MT);
        //ed += t2 -t1;

        // Allocation of new L matrix
        //UMat* LU = new UMat(IBlur->rows, IBlur->cols, IBlur->type());
        //t1 = get_time(TIMER_MT);
        Mat* L = new Mat(IBlur->rows, IBlur->cols, IBlur->type());
        //t2 = get_time(TIMER_MT);
        //allocation += t2-t1;
        //*L = IErode.getMat(ACCESS_READ) + IDilate.getMat(ACCESS_READ) - (2*(IBlur->getMat(ACCESS_READ)));

        // Add calculation on GPU
        //t1 = get_time(TIMER_MT);
        add(IErode, IDilate, IErode);
        add(IBlurU, IBlurU, IBlurU);
        //t2 = get_time(TIMER_MT);
        //calc += t2-t1;
        //subtract(IErode, IBlurU, *LU);
        
        // Copy back may be??
        //t1 = get_time(TIMER_MT);
        UMat LU = L->getUMat(ACCESS_WRITE);
        //t2 = get_time(TIMER_MT);
        //copyB += t2 - t1;

        // Sub on GPU
        //t1 = get_time(TIMER_MT);
        subtract(IErode, IBlurU, LU);
        //t2 = get_time(TIMER_MT);
        //sub += t2 - t1;

        //Mat* L = new Mat(LU->getMat(ACCESS_READ));
        //UMat* L = new UMat(IBlur->rows, IBlur->cols, IBlur->type());
        //Laplacian(*IBlur, *L, -1);
        // Pushing on buffer
        //t1 = get_time(TIMER_MT);
        out_L->push(L, id);
        //t2 = get_time(TIMER_MT);
        //push += t2 - t1;

        // Freeing memory
        //t1 = get_time(TIMER_MT);
        IErode.release();
        IDilate.release();
        IBlur->release();
        IBlurU.release();
        //t2 = get_time(TIMER_MT);
        release += t2 - t1;
        //LU->release();
        i++;
    }
    end = get_time(TIMER_THREAD);
    *time = (end-start)*1.0/BILLION;
    //*ctime = copy *1.0/BILLION;
    //cout << "\t\t\tGET\t"<< get *1.0/BILLION << "\n";
    //cout << "\t\t\tCOPY\t"<< copy *1.0/BILLION << "\n";
    //cout << "\t\t\tED\t"<< ed *1.0/BILLION << "\n";
    //cout << "\t\t\tALLOC\t"<< allocation *1.0/BILLION << "\n";
    //cout << "\t\t\tCALC\t"<< calc *1.0/BILLION << "\n";
    //cout << "\t\t\tCOPY-B\t"<< copyB *1.0/BILLION << "\n";
    //cout << "\t\t\tSUB\t"<< sub *1.0/BILLION << "\n";
    //cout << "\t\t\tPUSH\t"<< push *1.0/BILLION << "\n";
    //cout << "\t\t\tFREE\t"<< release *1.0/BILLION << "\n";
    DEBUG(cout << "Non-linear Laplacian Done\n");
}

// Gradient Node
void Gradient(CircularBuffer<Mat>* in_I, CircularBuffer<Mat>* out_dI, float* time) {
    Timestamp start, end;
    start = get_time(TIMER_THREAD);
    ocl::setUseOpenCL(OPENCL);
    DEBUG(cout << "Gradient Starts\n");
    string id = "G";
    int i = 0;
    while(i < NUM_FRAMES) {
        //waitKey(0);
        Mat* I = in_I->pop(id);
        DEBUG(cout << "[" << id << "]: Iteration " << i << "\n");
        // Gradient Computation
        //Mat dIx, dIy;
        UMat dIx, dIy;
        UMat IU = I->getUMat(ACCESS_READ);
        Sobel(IU, dIx, -1, 1, 0, 1);
        Sobel(IU, dIy, -1, 0, 1, 1);
        Mat* dI = new Mat(I->rows, I->cols, I->type());
        magnitude(dIx, dIy, dI->getUMat(ACCESS_WRITE));
        out_dI->push(dI, id);
        dIx.release();
        dIy.release();
        I->release();
        IU.release();
        i++;
    }
    end = get_time(TIMER_THREAD);
    *time = (end-start)*1.0/BILLION;
    DEBUG(cout << "Gradient Done\n");
}

// Find Zero Crossings Node
void ZeroCrossings(Mat SE, CircularBuffer<Mat>* in_L, CircularBuffer<Mat>* out_Z, float* time) {
    Timestamp start, end;
    start = get_time(TIMER_THREAD);
    ocl::setUseOpenCL(OPENCL);
    DEBUG(cout << "ZeroCrossings Start\n");
    string id = "ZC";
    int i = 0;
    while(i < NUM_FRAMES) {
        //waitKey(0);
        Mat* L = in_L->pop(id);
        DEBUG(cout << "[" << id << "]: Iteration " << i << "\n");

        Mat PErode, PDilate;
        Mat P;
        compare(*L, 0, P, CMP_GE);
        //Mat P = (*L >= 0);
        erode(P, PErode, SE);
        dilate(P, PDilate, SE);
        //Mat* Z = new Mat(L->rows, L->cols, L->type());
        //*Z = PDilate.getMat(ACCESS_READ)-PErode.getMat(ACCESS_READ);
        Mat* Z = new Mat(L->rows, L->cols, L->type());
        subtract(PDilate, PErode, *Z);
        //*Z = PDilate.getMat(ACCESS_READ)-PErode.getMat(ACCESS_READ);
        out_Z->push(Z, id);
        L->release();
        P.release();
        PDilate.release();
        PErode.release();
        i++;
    }
    end = get_time(TIMER_THREAD);
    *time = (end-start)*1.0/BILLION;
    DEBUG(cout << "ZeroCrossings Done\n");

}

// Reject Zero crossings Node
void RejectZeroCrossings(float Threshold, CircularBuffer<Mat>* in_dI,
                         CircularBuffer<Mat>* in_Z, CircularBuffer<Mat>* out_E, float* time) {
    Timestamp start, end;
    start = get_time(TIMER_THREAD);
    ocl::setUseOpenCL(OPENCL);
    // Reject Zero Crossings
    DEBUG(cout << "Reject Zero Crossings Starts\n");
    string id = "RZC";
    int i = 0;
    while(i < NUM_FRAMES) {
        //waitKey(0);
        Mat* dI = in_dI->pop(id);
        Mat* Z = in_Z->pop(id);
        DEBUG(cout << "[" << id << "]: Iteration " << i << "\n");
        double dI_max;
        Mat* E = new Mat(Z->rows, Z->cols, Z->type());
        minMaxLoc(*dI, NULL, &dI_max);

        //*E = *Z & (dI->getMat(ACCESS_READ)> Threshold*dI_max);
        Mat temp;
        compare(*dI, Threshold*dI_max, temp, CMP_GT);
        bitwise_and(*Z, temp, *E);
        out_E->push(E, id);
        dI->release();
        Z->release();
        temp.release();
        i++;
    }
    end = get_time(TIMER_THREAD);
    *time = (end-start)*1.0/BILLION;
    DEBUG(cout << "Reject Zero Crossings Done\n");
}

// Producer thread, feeds in the same image NUM_FRAMES times to Gaussian and
// Gradient node
void producer_thread(Mat* I, CircularBuffer<Mat>* out_I_Gaussian, CircularBuffer<Mat>* out_I_Gradient, float* time) {
    Timestamp start, end;
    start = get_time(TIMER_THREAD);
    ocl::setUseOpenCL(OPENCL);
    DEBUG(cout << "Producer Start\n");
    string id = "P";
    int i = 0;
    //Mat UI = I->getMat(ACCESS_READ);
    while(i < NUM_FRAMES) {
        //waitKey(0);
        DEBUG(cout << "[" << id << "]: Iteration " << i << "\n");
        Mat* I1 = new Mat(I->rows, I->cols, I->type());
        Mat* I2 = new Mat(I->rows, I->cols, I->type());
        I->copyTo(*I1);
        I->copyTo(*I2);
        out_I_Gaussian->push(I1, id);
        out_I_Gradient->push(I2, id);
        i++;
    }

    //UI.release();
    I->release();
    end = get_time(TIMER_THREAD);
    *time = (end-start)*1.0/BILLION;
    DEBUG(cout << "Producer Done!\n");
}

// Consumer Node for displaying the image
void consumer_thread(CircularBuffer<Mat>* in_E, float* time) {
    Timestamp start, end;
    start = get_time(TIMER_THREAD);
    ocl::setUseOpenCL(OPENCL);
    DEBUG(cout << "Consumer Start\n");
    string id = "C";
    int i = 0;
    while(i < NUM_FRAMES) {
        //waitKey(0);
        Mat* E = in_E->pop(id);
        //E->getMat(ACCESS_RW);
        DEBUG(cout << "[" << id << "]: Iteration " << i << "\n");
        //imshow(window_name, *E);
        //waitKey(0);
        E->release();
        i++;
    }
    end = get_time(TIMER_THREAD);
    *time = (end-start)*1.0/BILLION;
    DEBUG(cout << "Consumer Done!\n");
}


int main(int argc, char* argv[]) {
    int bufferSize = stoi(argv[1]);
    int iterations = 1;
    if(argc == 4)
      iterations = stoi(argv[3]);
    float Threshold = 0.1;
    float Sn = 1.0;

    DEBUG(cout << "Accelerating Using OpenCV" << CV_VERSION << "\n");
    ocl::setUseOpenCL(OPENCL);
    cout << "Use OpenCL: " << ocl::useOpenCL() << "\n";

    /// Load an image
    Timestamp start, end;
    //start = get_time(TIMER_MT);
    //end = get_time(TIMER_MT);
    //cout << "Running time  = " << (end-start * 1.0)/BILLION  << " seconds\n";
    Mat B = getStructuringElement(MORPH_CROSS, Size(3,3));

    namedWindow( window_name, CV_WINDOW_AUTOSIZE );
    DEBUG(cout << "Main Starts\n");

    float tGS, tG, tL, tRZC, tZC, tP, tC, tMain, tCopy;
    float aggtGS=0, aggtG=0, aggtL=0, aggtRZC=0, aggtZC=0, aggtP=0, aggtC=0, aggtMain=0;
    for(int i = 0; i<iterations+1; i++) {
        Mat src1 = readImage(argv[2]);
        //cout << "Image Size = " << src1.rows << ", " << src1.cols << "\n";
        start = get_time(TIMER_MT);
        // Producer
        CircularBuffer<Mat> Producer_Gaussian_I(bufferSize, "Producer-->Gaussian-I");
        CircularBuffer<Mat> Producer_Gradient_I(bufferSize, "Producer-->Gradient-I");
        thread producer(producer_thread, &src1, &Producer_Gaussian_I, &Producer_Gradient_I, &tP);

        // Gaussian Smoothening Node
        CircularBuffer<Mat> Gaussian_Laplacian_IBlur(bufferSize, "Gaussian-->Laplacian-IBlur");
        thread GaussianNode(GaussianSmoothening, Sn, &Producer_Gaussian_I, &Gaussian_Laplacian_IBlur, &tGS);

        // Gradient Node
        CircularBuffer<Mat> Gradient_Reject_dI(bufferSize, "Gradient-->Reject-dI");
        thread GradientNode(Gradient, &Producer_Gradient_I, &Gradient_Reject_dI, &tG);

        // Laplacian Node
        CircularBuffer<Mat> Laplacian_Zero_L(bufferSize, "Laplacian-->Zero-L");
        thread LaplacianNode(NonLinearLaplacian, B, &Gaussian_Laplacian_IBlur, &Laplacian_Zero_L, &tL, &tCopy);

        // Zero Crossings Node
        CircularBuffer<Mat> Zero_Reject_Z(bufferSize, "Zero-->Reject-Z");
        thread ZeroCrossingsNode(ZeroCrossings, B, &Laplacian_Zero_L, &Zero_Reject_Z, &tZC);

        // Reject Zero Crossings Node
        CircularBuffer<Mat> Reject_Consumer_E(bufferSize, "Reject-->Consumer-E");
        thread RejectZeroNode(RejectZeroCrossings, Threshold, &Gradient_Reject_dI, &Zero_Reject_Z, &Reject_Consumer_E, &tRZC);

        // Consumer
        thread consumer(consumer_thread, &Reject_Consumer_E, &tC);

        // Wait for threads to finish
        producer.join();
        GaussianNode.join();
        GradientNode.join();
        LaplacianNode.join();
        ZeroCrossingsNode.join();
        RejectZeroNode.join();
        consumer.join();

        end = get_time(TIMER_MT);
        tMain = (end-start*1.0)/BILLION;
        // Skip first iteration numbers to avoid first iteration which sometimes
        // get scheduled entirely on CPU
        if(i > 0) {
          aggtMain += tMain;
          aggtGS += tGS;
          aggtG += tG;
          aggtL += tL;
          aggtZC += tZC;
          aggtRZC += tRZC;
          aggtP += tP;
          aggtC += tC;
        }
        //cout << i << ":\t(Main)\t" << tMain << "\n";
        //cout << i << ":\t\t(L)\t" << tL << "\n";
        //cout << i << ":\t\t\t(Copy)\t"<< tCopy << "\n";
        //cout << i << ":\t\t(G)\t" << tG << "\n";
        //cout << i << ":\t\t(GS)\t" << tGS << "\n";
        //cout << i << ":\t\t(ZC)\t" << tZC << "\n";
    }
    cout << "Main Running time  = " << aggtMain/iterations << " seconds\n";
    cout << "GS Running time  = " << aggtGS/iterations  << " seconds\n";
    cout << "G Running time  = " << aggtG/iterations  << " seconds\n";
    cout << "L Running time  = " << aggtL/iterations  << " seconds\n";
    cout << "ZC Running time  = " << aggtZC/iterations  << " seconds\n";
    cout << "RZC Running time  = " << aggtRZC/iterations  << " seconds\n";
    cout << "P Running time  = " << aggtP/iterations  << " seconds\n";
    cout << "C Running time  = " << aggtC/iterations  << " seconds\n";

    cout << "Total Running time = " << (aggtGS+aggtG+aggtL+aggtZC+aggtRZC+aggtP+aggtC)/iterations << "\n";
    /// Create a window
    return 0;
}
