// Tile size
#define TILE_SIZE 16

// OpenCL Kernel for matrix multiply, C = A * B
// using shared memory
__kernel void matrixMul_sh(__global float* C,
                           __global float* A,
                           __global float* B,
                           int k,
                           int n) {

  int tx = get_local_id(0); //2D Local Thread ID x
  int ty = get_local_id(1); //2D Local Thread ID y

  int bx = get_group_id(0); //2D Block ID x
  int by = get_group_id(1); //2D Block ID y

  int col = bx * TILE_SIZE + tx;
  int row = by * TILE_SIZE + ty;

  // Static work-group (thread block) local (shared) memory allocations
  __local float A_s[TILE_SIZE][TILE_SIZE];
  __local float B_s[TILE_SIZE][TILE_SIZE];

  // Initialize accumulator
  float res = 0.0f;

  int i,l;

  for (l = 0; l < k/TILE_SIZE; l++) {
    // Transfer tiles of A and B to local (shared) memory
    A_s[ty][tx] = A[row*k + TILE_SIZE*l + tx];
    B_s[ty][tx] = B[(TILE_SIZE*l + ty)*n + col];

    // Barrier to synchronize all threads
    barrier(CLK_LOCAL_MEM_FENCE);
    // Now the local submatricies A_s and B_s are valid

    /* Multiply the two submatrices. Each thread computes one element of the *
     * block submatrix                                                       */
    for (i = 0; i < TILE_SIZE; i++)
      res += A_s[ty][i]*B_s[i][tx];

    /* Barrier: calculations must be completed before next memory transfer    *
     * can start                                                              */
    barrier(CLK_LOCAL_MEM_FENCE);

  }

  // Write in device memory
  C[row * n + col] = res;

}

