/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/
/*
  Implementing Breadth first search on CUDA using algorithm given in DAC'10
  paper "An Effective GPU Implementation of Breadth-First Search"

  Copyright (c) 2010 University of Illinois at Urbana-Champaign.
  All rights reserved.

  Permission to use, copy, modify and distribute this software and its documentation for
  educational purpose is hereby granted without fee, provided that the above copyright
  notice and this permission notice appear in all copies of this software and that you do
  not sell the software.

  THE SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY OF ANY KIND,EXPRESS, IMPLIED OR
  OTHERWISE.

  Author: Lijiuan Luo (lluo3@uiuc.edu)
  Revised for Parboil 2.5 Benchmark Suite by: Geng Daniel Liu (gengliu2@illinois.edu)
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "parboil.h"
#include "config.h"
#include <visc.h>

/**********
Define colors for BFS
1) the definition of White, gray and black comes from the text book "Introduction to Algorithms"
2) For path search problems, people may choose to use different colors to record the found paths.
Therefore we reserve numbers (0-16677216) for this purpose. Only nodes with colors bigger than
UP_LIMIT are free to visit 
3) We define two gray shades to differentiate between the new frontier nodes and the old frontier nodes that
 have not been marked BLACK 
*************/

//#pragma OPENCL EXTENSION cl_khr_global_int32_extended_atomics: enable
//#pragma OPENCL EXTENSION cl_khr_global_int32_base_atomics: enable
//#pragma OPENCL EXTENSION cl_khr_local_int32_base_atomics: enable
/*****************************************************************************
This is the  most general version of BFS kernel, i.e. no assumption about #block in the grid  
\param q1: the array to hold the current frontier
\param q2: the array to hold the new frontier
\param g_graph_nodes: the nodes in the input graph
\param g_graph_edges: the edges i nthe input graph
\param g_color: the colors of nodes
\param g_cost: the costs of nodes
\param no_of_nodes: the number of nodes in the current frontier
\param tail: pointer to the location of the tail of the new frontier. *tail is the size of the new frontier 
\param gray_shade: the shade of the gray in current BFS propagation. See GRAY0, GRAY1 macro definitions for more details
\param k: the level of current propagation in the BFS tree. k= 0 for the first propagation.
***********************************************************************/

//typedef struct {
//} VoidRetTy;

typedef struct __attribute__((__packed__)) {
  int* q1; size_t bytesq1; 
  int* q2; size_t bytesq2; 
  struct Node* graph_nodes; size_t bytes_graph_nodes; 
  struct Edge* graph_edges; size_t bytes_graph_edges;
  int* color; size_t bytes_color;
  int* cost; size_t bytes_cost;
  int* tail; size_t bytes_tail;
  int no_of_nodes;
  int gray_shade;
  int k;
  long block;
  long grid;
  //VoidRetTy* out;
} RootIn;

void packData(RootIn* args,
              int* q1, size_t bytesq1, 
              int* q2, size_t bytesq2, 
              struct Node* graph_nodes, size_t bytes_graph_nodes, 
              struct Edge* graph_edges, size_t bytes_graph_edges,
              int* color, size_t bytes_color,
              int* cost, size_t bytes_cost,
              int* tail, size_t bytes_tail,
              int no_of_nodes,
              int gray_shade,
              int k,
              long block,
              long grid) {
  args->q1 = q1;
  args->bytesq1 = bytesq1;
  args->q2 = q2;
  args->bytesq2 = bytesq2;
  args->graph_nodes = graph_nodes;
  args->bytes_graph_nodes = bytes_graph_nodes;
  args->graph_edges = graph_edges;
  args->bytes_graph_edges = bytes_graph_edges;
  args->color = color;
  args->bytes_color = bytes_color;
  args->cost = cost;
  args->bytes_cost = bytes_cost;
  args->tail = tail;
  args->bytes_tail = bytes_tail;
  args->no_of_nodes = no_of_nodes;
  args->gray_shade = gray_shade;
  args->k = k;
  args->block = block;
  args->grid = grid;
}

void Allocation(long block) {
  // Memory shared between threadblocks
  void* local_q_tail = __visc__malloc(sizeof(int));
  void* local_q = __visc__malloc(LOCAL_MEM_SIZE*sizeof(int));
  void* shift = __visc__malloc(sizeof(int));
  
  __visc__return(6, local_q_tail, sizeof(int), local_q, LOCAL_MEM_SIZE*sizeof(int), shift, sizeof(int));
}

//VoidRetTy
void
BFSLeaf(int *q1, size_t bytesq1, 
           int *q2, size_t bytesq2,
           struct Node *g_graph_nodes, size_t bytesg_graph_nodes,
           struct Edge *g_graph_edges, size_t bytesg_graph_edges,
           int *g_color, size_t bytesg_color,
           int *g_cost, size_t bytesg_cost,
           int *tail, size_t bytestail,
           int no_of_nodes,
           int gray_shade, 
           int k,
           // data local to thread block. The next three arguments should
           // ideally be placed in local memory
           int *local_q_tail, size_t byteslocal_q_tail, 
           int *local_q, size_t byteslocal_q,
           int *shift, size_t bytesshift
           ) 
{

  __visc__hint(visc::DEVICE);
  __visc__attributes(6, q1, g_graph_nodes, g_graph_edges, g_color, g_cost, tail,
                      4, q2, g_color, g_cost, tail);

  void* thisNode = __visc__getNode();
  void* parentNode = __visc__getParentNode(thisNode);
  int lx = __visc__getNodeInstanceID_x(thisNode);
  int gx = __visc__getNodeInstanceID_x(parentNode);
  int dimx = __visc__getNumNodeInstances_x(thisNode);

  if(lx == 0){
    *local_q_tail = 0;//initialize the tail of w-queue
  }

  __visc__barrier(); 

  //first, propagate and add the new frontier elements into w-queues
  //int tid = get_group_id(0)*MAX_THREADS_PER_BLOCK + get_local_id(0);
  int tid = gx * dimx + lx;

  if( tid<no_of_nodes)
  {
    int pid = q1[tid]; //the current frontier node, or the parent node of the new frontier nodes 
    g_color[pid] = BLACK;
    int cur_cost = g_cost[pid];
    //into
    struct Node cur_node = g_graph_nodes[pid];

    for(int i=cur_node.x; i<cur_node.y + cur_node.x; i++)//visit each neighbor of the
      //current frontier node.
    {
      struct Edge cur_edge = g_graph_edges[i];
      int id = cur_edge.x;
      int cost = cur_edge.y;
      cost += cur_cost;

      int orig_cost = __visc__atomic_min(&g_cost[id],cost);
      if(orig_cost > cost){//the node should be visited
        if(g_color[id] > UP_LIMIT){
         int old_color = __visc__atomic_xchg(&g_color[id],gray_shade);
          //this guarantees that only one thread will push this node
          //into a queue
          if(old_color != gray_shade) {
            //atomic operation guarantees the correctness
            //even if multiple warps are executing simultaneously
            int index = __visc__atomic_add(local_q_tail, 1);
            local_q[index] = id;
          }
        }
      }
    }
  }
  
  __visc__barrier(); 

  if(lx == 0){
    int tot_sum = *local_q_tail; 
    //the offset or "shift" of the block-level queue within the grid-level queue
    //is determined by atomic operation
    *shift = __visc__atomic_add (tail,tot_sum);
  }
  
  __visc__barrier(); 

  //shift within a w-queue
  int local_shift = lx;

  while(local_shift < *local_q_tail){
    q2[*shift + local_shift] = local_q[local_shift];
    //multiple threads are copying elements at the same time,
    //so we shift by multiple elements for next iteration  
    local_shift += dimx;
  }
}

//VoidRetTy
void BlockingBFS(int *q1, size_t bytesq1, 
                int *q2, size_t bytesq2,
                struct Node *g_graph_nodes, size_t bytesg_graph_nodes,
                struct Edge *g_graph_edges, size_t bytesg_graph_edges,
                int *g_color, size_t bytesg_color,
                int *g_cost, size_t bytesg_cost,
                int *tail, size_t bytestail,
                int no_of_nodes,
                int gray_shade, 
                int k,
                long block,
                // data local to thread block. The next three arguments should
                // ideally be placed in local memory
                int *local_q_tail, size_t byteslocal_q_tail, 
                int *local_q, size_t byteslocal_q,
                int *shift, size_t bytesshift) {
  __visc__hint(visc::CPU_TARGET);
  __visc__attributes(6, q1, g_graph_nodes, g_graph_edges, g_color, g_cost, tail,
                      4, q2, g_color, g_cost, tail);

  void* AllocationNode = __visc__createNodeND(0, Allocation);
  void* BFSLeafNode = __visc__createNodeND(1, BFSLeaf, block);

  // Bind edges
  __visc__bindIn(AllocationNode, 17, 0, 0); // Bind block
  __visc__bindIn(BFSLeafNode, 0, 0, 0); // Bind q1
  __visc__bindIn(BFSLeafNode, 1, 1, 0); // Bind bytes_q1
  __visc__bindIn(BFSLeafNode, 2, 2, 0); // Bind q2
  __visc__bindIn(BFSLeafNode, 3, 3, 0); // Bind bytes_q2 
  __visc__bindIn(BFSLeafNode, 4, 4, 0); // Bind graph_nodes
  __visc__bindIn(BFSLeafNode, 5, 5, 0); // Bind bytes_graph_nodes
  __visc__bindIn(BFSLeafNode, 6, 6, 0); // Bind graph_edges
  __visc__bindIn(BFSLeafNode, 7, 7, 0); // Bind bytes_graph_edges
  __visc__bindIn(BFSLeafNode, 8, 8, 0); // Bind color
  __visc__bindIn(BFSLeafNode, 9, 9, 0); // Bind bytes_color
  __visc__bindIn(BFSLeafNode, 10, 10, 0); // Bind cost
  __visc__bindIn(BFSLeafNode, 11, 11, 0); // Bind bytes_cost
  __visc__bindIn(BFSLeafNode, 12, 12, 0); // Bind tail
  __visc__bindIn(BFSLeafNode, 13, 13, 0); // Bind bytes_tail
  __visc__bindIn(BFSLeafNode, 14, 14, 0); // Bind no_of_nodes
  __visc__bindIn(BFSLeafNode, 15, 15, 0); // Bind gray_shade
  __visc__bindIn(BFSLeafNode, 16, 16, 0); // Bind k
  
  // Create Edges between AllocationNode and BFSLeafNodeNode
  __visc__edge(AllocationNode, BFSLeafNode, 1, 0, 17, 0); // Edge local_q_tail 
  __visc__edge(AllocationNode, BFSLeafNode, 1, 1, 18, 0); // Edge bytes_local_q_tail 
  __visc__edge(AllocationNode, BFSLeafNode, 1, 2, 19, 0); // Edge local_q 
  __visc__edge(AllocationNode, BFSLeafNode, 1, 3, 20, 0); // Edge bytes_local_q 
  __visc__edge(AllocationNode, BFSLeafNode, 1, 4, 21, 0); // Edge shift 
  __visc__edge(AllocationNode, BFSLeafNode, 1, 5, 22, 0); // Edge bytes_shift 
}

//VoidRetTy
void BFS_Root(int *q1, size_t bytesq1, 
              int *q2, size_t bytesq2,
              struct Node *g_graph_nodes, size_t bytesg_graph_nodes,
              struct Edge *g_graph_edges, size_t bytesg_graph_edges,
              int *g_color, size_t bytesg_color,
              int *g_cost, size_t bytesg_cost,
              int *tail, size_t bytestail,
              int no_of_nodes,
              int gray_shade, 
              int k,
              long block,
              long grid) {
  __visc__hint(visc::CPU_TARGET);
  __visc__attributes( 6, q1, g_graph_nodes, g_graph_edges, g_color, g_cost, tail,
                      4, q2, g_color, g_cost, tail);
  void* BlockingBFSNode = __visc__createNodeND(1, BlockingBFS, grid);

  // Bind edges
  __visc__bindIn(BlockingBFSNode, 0, 0, 0); // Bind q1
  __visc__bindIn(BlockingBFSNode, 1, 1, 0); // Bind bytes_q1
  __visc__bindIn(BlockingBFSNode, 2, 2, 0); // Bind q2
  __visc__bindIn(BlockingBFSNode, 3, 3, 0); // Bind bytes_q2 
  __visc__bindIn(BlockingBFSNode, 4, 4, 0); // Bind graph_nodes
  __visc__bindIn(BlockingBFSNode, 5, 5, 0); // Bind bytes_graph_nodes
  __visc__bindIn(BlockingBFSNode, 6, 6, 0); // Bind graph_edges
  __visc__bindIn(BlockingBFSNode, 7, 7, 0); // Bind bytes_graph_edges
  __visc__bindIn(BlockingBFSNode, 8, 8, 0); // Bind color
  __visc__bindIn(BlockingBFSNode, 9, 9, 0); // Bind bytes_color
  __visc__bindIn(BlockingBFSNode, 10, 10, 0); // Bind cost
  __visc__bindIn(BlockingBFSNode, 11, 11, 0); // Bind bytes_cost
  __visc__bindIn(BlockingBFSNode, 12, 12, 0); // Bind tail
  __visc__bindIn(BlockingBFSNode, 13, 13, 0); // Bind bytes_tail
  __visc__bindIn(BlockingBFSNode, 14, 14, 0); // Bind no_of_nodes
  __visc__bindIn(BlockingBFSNode, 15, 15, 0); // Bind gray_shade
  __visc__bindIn(BlockingBFSNode, 16, 16, 0); // Bind k
  __visc__bindIn(BlockingBFSNode, 17, 17, 0); // Bind block

  
}



FILE *fp;
char* readFile(const char* fileName)
{
  FILE* fp;
  fp = fopen(fileName,"r");
  if(fp == NULL)
  {
    printf("Error 1!\n");
    exit(1);
  }

  fseek(fp,0,SEEK_END);
  long size = ftell(fp);
  rewind(fp);

  char* buffer = (char*)malloc(sizeof(char)*size);
  if(buffer  == NULL)
  {
    printf("Error 2!\n");
    fclose(fp);
    exit(1);
  }
  size_t res = fread(buffer,1,size,fp);
  if(res != size)
  {
    printf("Error 3!\n");
    fclose(fp);
    exit(1);
  }

  fclose(fp);
  return buffer;
}
const int h_top = 1;
const int zero = 0;
void runGPU(int argc, char** argv);
////////////////////////////////////////////////////////////////////////////////
// Main Program
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv)
{

  //the number of nodes in the graph
  int num_of_nodes = 0;
  //the number of edges in the graph
  int num_of_edges = 0;
  struct pb_Parameters *params;
  struct pb_TimerSet timers;

  params = pb_ReadParameters(&argc, argv);
  if ((params->inpFiles[0] == NULL) || (params->inpFiles[1] != NULL))
  {
    fprintf(stderr, "Expecting one input filename\n");
    exit(-1);
  }

  //Read in Graph from a file
  fp = fopen(params->inpFiles[0],"r");
  if(!fp)
  {
    printf("Error Reading graph file\n");
    return 0;
  }
  int source;

  fscanf(fp,"%d",&num_of_nodes);
  // allocate host memory
  struct Node* graph_nodes = (struct Node*) malloc(sizeof(struct Node)*num_of_nodes);
  int *color = (int*) malloc(sizeof(int)*num_of_nodes);
  int start, edgeno;
  // initalize the memory
  int i;
  for( i = 0; i < num_of_nodes; i++)
  {
    fscanf(fp,"%d %d",&start,&edgeno);
    graph_nodes[i].x = start;
    graph_nodes[i].y = edgeno;
    color[i]=WHITE;
  }
  //read the source node from the file
  fscanf(fp,"%d",&source);
  fscanf(fp,"%d",&num_of_edges);
  int id,edge_cost;
  struct Edge* graph_edges = (struct Edge*) malloc(sizeof(struct Edge)*num_of_edges);
  for(i=0; i < num_of_edges ; i++)
  {
    fscanf(fp,"%d",&id);
    fscanf(fp,"%d",&edge_cost);
    graph_edges[i].x = id;
    graph_edges[i].y = edge_cost;
  }
  if(fp)
    fclose(fp);

  pb_InitializeTimerSet(&timers);
  __visc__init();

  pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);
  // allocate mem for the result on host side
  int* cost = (int*) malloc( sizeof(int)*num_of_nodes);
  for(i = 0; i < num_of_nodes; i++){
    cost[i] = INF;
  }
  cost[source] = 0;

  size_t bytes_graph_nodes = num_of_nodes* sizeof(struct Node);
  size_t bytes_graph_edges = num_of_edges* sizeof(struct Edge);
  size_t bytes_cost = sizeof(int) * num_of_nodes; 

  int* q1 = (int*) malloc(sizeof(int)*num_of_nodes);
  int* q2 = (int*) malloc(sizeof(int)*num_of_nodes);
  int* tail = (int*) malloc(sizeof(int));

  llvm_visc_track_mem(graph_nodes, bytes_graph_nodes);
  llvm_visc_track_mem(graph_edges, bytes_graph_edges);
  llvm_visc_track_mem(cost, bytes_cost);
  llvm_visc_track_mem(color, bytes_cost);
  // Allocating stuff on host side, but these can also be allocated in the graph
  llvm_visc_track_mem(q1, bytes_cost);
  llvm_visc_track_mem(q2, bytes_cost);
  // Scalar variable read/written by both graph and host.
  llvm_visc_track_mem(tail, sizeof(int));

  int num_of_blocks;
  int num_of_threads_per_block;

  // Initializations. Can some of these be done in the graph. That way we can
  // move these arrays completely in the graph
  *tail = h_top;
  // Potential source of inefficiency. 
  //Entire array would be copied intially
  cost[0] = zero;
  q1[0] = source;

  int num_t;//number of threads
  int k=0;//BFS level index
  int gray;

  long grid = num_of_blocks;
  long block = num_of_threads_per_block;
  // Pack data in struct
  RootIn* args = (RootIn*) malloc(sizeof(RootIn));
  packData(args, 
          q1, bytes_cost, 
          q2, bytes_cost,
          graph_nodes, bytes_graph_nodes,
          graph_edges, bytes_graph_edges,
          color, bytes_cost,
          cost, bytes_cost,
          tail, sizeof(int),
          num_of_nodes,
          gray, 
          k,
          block,
          grid
          );


  pb_SwitchToTimer(&timers, visc_TimerID_COMPUTATION);
  do
  {
    llvm_visc_request_mem(tail, sizeof(int));
    num_t = *tail;
    //printf("tail for iteration %d = %d\n",k, num_t);
    *tail = 0;
    //tail = 0;

    if(num_t == 0){//frontier is empty
      break;
    }

    num_of_blocks = (int)ceil(num_t/(double)MAX_THREADS_PER_BLOCK);
    num_of_threads_per_block = num_t > MAX_THREADS_PER_BLOCK ? MAX_THREADS_PER_BLOCK : num_t;

    args->grid = num_of_blocks;
    args->block = num_of_threads_per_block;

    args->no_of_nodes = num_t;
    args->k = k;

    if(k%2 == 0){
      args->gray_shade = GRAY0;
    }
    else{
      args->gray_shade = GRAY1;
    }
    //void* bfsDFG = __visc__node(BFS_kernel, 2, 1, block, grid, 17, 
                                //q1, bytes_cost, 
                                //q2, bytes_cost,
                                //graph_nodes, bytes_graph_nodes,
                                //graph_edges, bytes_graph_edges,
                                //color, bytes_cost,
                                //cost, bytes_cost,
                                //tail, sizeof(int),
                                //num_of_nodes,
                                //gray, 
                                //k,
                                //0);
    void* bfsDFG = __visc__launch(0, BFS_Root, (void*) args);
    __visc__wait(bfsDFG);
    // Swap q1 and q2
    // Swap q1 and q2
    int* temp = args->q1;
    args->q1 = args->q2;
    args->q2 = temp;
    k++;
  } while(1);

  // copy result from device to host
  pb_SwitchToTimer(&timers, pb_TimerID_COPY);

  llvm_visc_request_mem(cost, bytes_cost);
  llvm_visc_request_mem(color, bytes_cost);

  pb_SwitchToTimer(&timers, pb_TimerID_COMPUTE);

  llvm_visc_untrack_mem(graph_nodes);
  llvm_visc_untrack_mem(graph_edges);
  llvm_visc_untrack_mem(cost);
  llvm_visc_untrack_mem(color);
  llvm_visc_untrack_mem(q1);
  llvm_visc_untrack_mem(q2);
  llvm_visc_untrack_mem(tail);

  pb_SwitchToTimer(&timers, pb_TimerID_NONE);
  pb_PrintTimerSet(&timers);
  __visc__cleanup();

  //Store the result into a file
  //FIXME: color is not even printed. Why are we reading it back??
  FILE *fp = fopen(params->outFile,"w");
  fprintf(fp, "%d\n", num_of_nodes);
  int j = 0;
  for(j=0;j<num_of_nodes;j++)
    fprintf(fp,"%d %d\n",j,cost[j]);
  fclose(fp);
  // cleanup memory
  free(graph_nodes);
  free(graph_edges);
  free(color);
  free(cost);
  free(q1);
  free(q2);


  pb_FreeParameters(params);
  return 0;
}
