/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

/*
 * Main entry of dense matrix-matrix multiplication kernel
 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <malloc.h>
#include <vector>
#include <iostream>
#include <parboil.h>
#include <visc.h>

// I/O routines
extern bool readColMajorMatrixFile(const char *fn, int &nr_row, int &nr_col, std::vector<float>&v);
extern bool writeColMajorMatrixFile(const char *fn, int, int, std::vector<float>&);
extern char* readFile(const char*);

// Parameters of tile sizes
#define TILE_SZ 16

#define CHECK_ERROR(errorMessage)           \
  if(clStatus != CL_SUCCESS)                \
  {                                         \
     std::cout<<errorMessage<<" Error!\n";  \
     std::cout<<"Line: "<<__LINE__<<"\n";   \
     exit(1);                               \
  }

void mysgemmNT( float* A, int lda, float* B, int ldb, float* C, int ldc, int k, float alpha, float beta )
{
    __visc__attributes(3, A, B, C, 1, C);
    float c0, c1, c2, c3;
    c0 = c1 = c2 = c3 = 0.0f;
    int m = 4 * get_global_id(0);
    int n = get_global_id(1);

    for (int i = 0; i < k; ++i) {
        float a0 = A[m + i * lda]; 
        float a1 = A[m + 1 + i * lda]; 
        float a2 = A[m + 2 + i * lda]; 
        float a3 = A[m + 3 + i * lda]; 

        float b = B[n + i * ldb];

        c0 += a0 * b;
        c1 += a1 * b;
        c2 += a2 * b;
        c3 += a3 * b;
    }
    C[m+n*ldc] = C[m+n*ldc] * beta + alpha * c0;
    C[m+1+n*ldc] = C[m+1+n*ldc] * beta + alpha * c1;
    C[m+2+n*ldc] = C[m+2+n*ldc] * beta + alpha * c2;
    C[m+3+n*ldc] = C[m+3+n*ldc] * beta + alpha * c3;
}

__attribute__((noinline)) void basicSgemm( char transa, char transb, int m, int n, int k, float alpha, float* A, size_t bytesA, int lda, float* B, size_t bytesB, int ldb, float beta, float* C, size_t bytesC, int ldc )
{
    if ((transa != 'N') && (transa != 'n')) {
        std::cerr << "unsupported value of 'transa' in regtileSgemm()" << std::endl;
        return;
    }

    if ((transb != 'T') && (transb != 't')) {
        std::cerr << "unsupported value of 'transb' in regtileSgemm()" << std::endl;
        return;
    }

    // In this code we assume the matrix sizes are multiple of tile size
    if ((m%TILE_SZ) || (n%TILE_SZ)) {
        std::cerr << "unsupported size of matrix. m should be multiple of " << TILE_SZ
                  << "; n should be multiple of " << TILE_SZ << std::endl;
    }

    unsigned db[2] = {TILE_SZ/4,TILE_SZ};
    unsigned dg[2] = {m/TILE_SZ*db[0],n/TILE_SZ*db[1]};

    unsigned sgemmDFG = __visc__node(mysgemmNT, 2, 2, db[0], db[1], dg[0]/db[0], dg[1]/db[1], 12, A, bytesA, lda, B, bytesB, ldb, C, bytesC, ldc, k, alpha, beta, 0);
    __visc__wait(sgemmDFG);
}

int main (int argc, char *argv[]) {

    struct pb_Parameters *params;
    struct pb_TimerSet timers;

    size_t A_sz, B_sz, C_sz;
    int matArow, matAcol;
    int matBrow, matBcol;
    std::vector<float> matA, matBT;

    pb_InitializeTimerSet(&timers);
    __visc__init();

    /* Read command line. Expect 3 inputs: A, B and B^T
       in column-major layout*/
    params = pb_ReadParameters(&argc, argv);
    if ((params->inpFiles[0] == NULL)
            || (params->inpFiles[1] == NULL)
            || (params->inpFiles[2] == NULL)
            || (params->inpFiles[3] != NULL))
    {
        fprintf(stderr, "Expecting three input filenames\n");
        exit(-1);
    }

    /* Read in data */
    pb_SwitchToTimer(&timers, pb_TimerID_IO);

    // load A
    readColMajorMatrixFile(params->inpFiles[0],
                           matArow, matAcol, matA);

    // load B^T
    readColMajorMatrixFile(params->inpFiles[2],
                           matBcol, matBrow, matBT);

    pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );
    // copy A to device memory
    A_sz = matArow*matAcol*sizeof(float);
    B_sz = matBrow*matBcol*sizeof(float);

    // allocate space for C
    C_sz = matArow*matBcol*sizeof(float);

    // OpenCL memory allocation
    std::vector<float> matC(matArow*matBcol);

    llvm_visc_track_mem(&matA.front(), A_sz);
    llvm_visc_track_mem(&matBT.front(), B_sz);
    llvm_visc_track_mem(&matC.front(), C_sz);
    // Copy A and B^T into device memory
    pb_SwitchToTimer( &timers, pb_TimerID_COMPUTE );

    for(size_t i=0; i<matC.size(); i++)
        matC[i] = 0.0f;

    pb_SwitchToTimer( &timers, pb_TimerID_NONE );

    // Use standard sgemm interface
    basicSgemm('N', 'T', matArow, matBcol, matAcol, 1.0f, \
               &matA.front(), A_sz, matArow, &matBT.front(), B_sz, matBcol, 0.0f, &matC.front(), C_sz, matArow);

    if (params->outFile) {
        pb_SwitchToTimer( &timers, pb_TimerID_COPY );

        /* Write C to file */
        llvm_visc_request_mem(&matC.front(), C_sz);
        pb_SwitchToTimer(&timers, pb_TimerID_IO);
        writeColMajorMatrixFile(params->outFile,
                                matArow, matBcol, matC);
    }

    pb_SwitchToTimer( &timers, visc_TimerID_MEM_UNTRACK );
    llvm_visc_untrack_mem(&matA.front());
    llvm_visc_untrack_mem(&matBT.front());
    llvm_visc_untrack_mem(&matC.front());
    pb_SwitchToTimer(&timers, pb_TimerID_NONE);

    double GPUtime = pb_GetElapsedTime(&(timers.timers[pb_TimerID_KERNEL]));
    std::cout<< "GFLOPs = " << 2.* matArow * matBcol * matAcol/GPUtime/1e9 << std::endl;
    pb_PrintTimerSet(&timers);
    __visc__cleanup();
    pb_FreeParameters(params);

    return 0;
}
