# (c) 2006 The Board of Trustees of the University of Illinois.

# This module takes care of parsing options for the Parboil driver.
#
# The main option parsing routine is parse_options().  Option parsing
# may print messages and terminate the program, but should not cause
# any other action to be taken.
import sys
from sys import stdout
from optparse import OptionParser
import os
import tempfile
import time
import datetime
import StringIO
import re
import actions
import csv
import copy
import globals

def invalid_option_message(progname, cmd, args):
    print "Unrecognized command '" + cmd + "'"
    print "'" + progname + " help' for options"

# Parsers for each mode.  These take the command line as parameters
# and return an OptionGetter.
#
# The 'help' field prints a help message.
#
# The 'run' field does the actual parsing.  If there is an error in
# the command line, it prints an error message and returns None;
# otherwise, it returns an action which will carry out the commands.

class OptionGetter:
    def __init__(self, help, run):
        self.help = help
        self.run = run

def help_options(progname, cmd, args):
    help_string = "usage: " + progname + " help [COMMAND]\nWithout parameters: list commands\nWith a parameter: Get help on COMMAND\n"
    get_help = lambda: stdout.write(help_string)

    def run():
        if args:
            try: helpcmd = parse_mode_options[args[0]]
            except KeyError:
                print "No help available for unrecognized command '" + args[0] + "'"
                return None

            helpcmd(progname, cmd, args).help()
        else:
            print "Commands: "
            print "  help      Display this help message"
            print "  list      List benchmarks"
            print "  describe  Show details on a benchmark"
            print "  clean     Clean up generated files in a benchmark"
            print "  compile   Compile a benchmark"
            print "  run       Run a benchmark"
            print ""
            print "To get help on a command: " + progname + " help COMMAND"

        return None

    return OptionGetter(get_help, run)

def list_options(progname, cmd, args):
    help_string = "usage: " + progname + " list\nList available benchmarks\n"
    get_help = lambda: stdout.write(help_string)

    def run():
        if args:
            print "Unexpected parameter or option after 'list'"
            return None
        else:
            return actions.list_benchmarks

    return OptionGetter(get_help, run)

def describe_options(progname, cmd, args):
    usage_string = progname + " describe [BENCHMARK]\nWithout parameters: describe all benchmarks in detail\nWith a parameter: describe BENCHMARK in detail"
    parser = OptionParser(usage=usage_string)

    def run():
        (opts, pos) = parser.parse_args(args)
        if len(pos) > 1:
            print "Too many parameters after 'describe'"
            return None
        elif len(pos) == 0:
            return actions.describe_benchmarks
        else:
            bmkname = pos[0]
            return lambda: actions.with_benchmark_named(bmkname,
                                                        actions.describe_benchmark)

    return OptionGetter(parser.print_help, run)

def clean_options(progname, cmd, args):
    usage_string = progname + " clean BENCHMARK [VERSION]\nDelete the object code and executable of BENCHMARK version VERSION;\nif no version is given, remove the object code and executable of all versions"
    parser = OptionParser(usage=usage_string)
    parser.add_option('-v', "--verbose",
                      action="store_true", dest="verbose", default=False,
                      help="Produce verbose status messages")

    def run():
        (opts, pos) = parser.parse_args(args)
        globals.verbose = opts.verbose

        if len(pos) == 0:
            print "Expecting two or three parameters after 'clean'"
            return None
        elif len(pos) == 1:
            bmkname = pos[0]
            return lambda: actions.with_benchmark_named(bmkname, actions.clean_benchmark)
        elif len(pos) == 2:
            bmkname = pos[0]
            ver = pos[1]
            return lambda: actions.with_benchmark_named(bmkname, lambda b: actions.clean_benchmark(b, ver))
        elif len(pos) == 3:
            bmkname = pos[0]
            ver = pos[1]
            platform = pos[2]
            return lambda: actions.with_benchmark_named(bmkname, lambda b: actions.clean_benchmark(b, ver, platform))
        else:
            print "Too many parameters after 'clean'"
            return None

    return OptionGetter(parser.print_help, run)

def compile_options(progname, cmd, args):
    help_string = "usage :" + progname + " compile BENCHMARK VERSION [PLATFORM=default]\nCompile version VERSION of BENCHMARK [for platform PLATFORM]"
    parser = OptionParser(usage = help_string)
    parser.add_option('-v', "--verbose",
                      action="store_true", dest="verbose", default=False,
                      help="Produce verbose status messages")

    def run():
        (opts, pos) = parser.parse_args(args)
        globals.verbose = opts.verbose

        if not len(pos) in [2, 3]:
            print "Expecting two or three parameters after 'compile'"
            return None
        else:
            bmkname = pos[0]
            ver = pos[1]
            try:
              platform = pos[2]
            except:
              platform = None
            return lambda: actions.with_benchmark_named(bmkname, lambda b: actions.compile_benchmark(b, ver, platform))

    return OptionGetter(parser.print_help, run)

def run_options(progname, cmd, args):
    usage_string = progname + " run BENCHMARK VERSION INPUT [PLATFORM=default]\nRun version VERSION of BENCHMARK with platform PLATFORM and data set INPUT"
    parser = OptionParser(usage=usage_string)
    parser.add_option('-C', "--no-check",
                      action="store_false", dest="check", default=True,
                      help="Skip the output check for this benchmark")
    parser.add_option('-v', "--verbose",
                      action="store_true", dest="verbose", default=False,
                      help="Produce verbose status messages")

    def run():
        (opts, pos) = parser.parse_args(args)
        globals.verbose = opts.verbose
        if not len(pos) in [3, 4]:
            print "Expecting three or four parameters after 'run'"
            return None
        else:
            bmkname = pos[0]
            ver = pos[1]
            inp = pos[2]
            try:
                platform = pos[3]
            except:
                platform = None
            ck=opts.check
            extra_opts = []
            return lambda: actions.with_benchmark_named(bmkname, lambda b: actions.run_benchmark(b, ver, inp, check=ck, extra_opts=extra_opts, platform=platform))

    return OptionGetter(parser.print_help, run)

def debug_options(progname, cmd, args):
    usage_string = progname + " debug BENCHMARK VERSION INPUT\nRun version VERSION of BENCHMARK with data set INPUT in debugger"
    parser = OptionParser(usage=usage_string)
    parser.add_option('-C', "--no-check",
                      action="store_false", dest="check", default=True,
                      help="Skip the output check for this benchmark")
    parser.add_option('-v', "--verbose",
                      action="store_true", dest="verbose", default=False,
                      help="Produce verbose status messages")

    def run():
        (opts, pos) = parser.parse_args(args)
        globals.verbose = opts.verbose
        if not len(pos) in [3, 4]:
            print "Expecting three or four parameters after 'run'"
            return None
        else:
            bmkname = pos[0]
            ver = pos[1]
            inp = pos[2]
            try:
                platform = pos[3]
            except:
                platform = None
            ck=opts.check
            extra_opts = []
            return lambda: actions.with_benchmark_named(bmkname, lambda b: actions.debug_benchmark(b, ver, inp, check=ck, extra_opts=extra_opts, platform=platform))

    return OptionGetter(parser.print_help, run)

def time_options(progname, cmd, args):
    globals.log = True
    usage_string = progname + " time [-v]"
    parser = OptionParser(usage = usage_string)
    #parser.add_option('-C', "--no-check",
    #                  action="store_false", dest="check", default=True,
    #                  help="Skip the output check for this benchmark")
    parser.add_option('-v', "--verbose",
                      action="store_true", dest="verbose", default=False,
                      help="Produce verbose status messages")

    def getLogDirname():
        ts = time.time()
        datestring = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d-%H:%M:%S')
        return "LOG-"+datestring

    def clearLog():
        globals.outlog = ""
        globals.errlog = ""

    def dump(data, filename):
        f = open(filename, "w")
        f.write(data)

    def dump_cleanLog(dirname):
        dump(globals.outlog, dirname+"/clean-stdout")
        dump(globals.errlog, dirname+"/clean-stderr")
        clearLog()

    def dump_buildLog(dirname):
        dump(globals.outlog, dirname+"/build-stdout")
        dump(globals.errlog, dirname+"/build-stderr")
        clearLog()

    def dump_runLog(dirname,test):
        dump(globals.outlog, dirname+"/run-"+test+"-stdout")
        dump(globals.errlog, dirname+"/run-"+test+"-stderr")
        clearLog()

    def addTime(timerlist, timings):
        val = 0.0
        for (label, category) in timerlist:
            val += float(timings[label][category])
        return val

    def compileTimings(timings):
        label_f = 'Final'
        label_pb = 'Parboil'
        if len(timings) >= 4:
            label_ptx = 'NVPTX_Timer'
            #label_ptx = 'SPIR_Timer'
            label_x86 = 'X86_Timer'
            label_gen = 'GenVISC_Timer'
            timings[label_f] = {}
            timings[label_f]['IO'] = addTime([(label_pb, 'IO')], timings)
            timings[label_f]['Memory Track'] = addTime([(label_pb, 'Mem_Track')], timings)
            timings[label_f]['Marshal Arguments'] = addTime([(label_gen, 'Arg_Pack')], timings)
            timings[label_f]['Pthread Create'] = addTime([(label_x86, 'Pthread_Create')], timings)
            timings[label_f]['Argument Unpack'] = addTime([(label_x86, 'Arg_Unpack')], timings)
            timings[label_f]['Total GPU Computation'] = addTime([(label_x86, 'Computation')], timings)
            timings[label_f]['Output Pack'] = addTime([(label_x86, 'Output_Pack')], timings)
            timings[label_f]['Output Unpack'] = addTime([(label_x86, 'Output_Unpack')], timings)
            timings[label_f]['Initialize Context'] = addTime([(label_ptx, 'Init_Ctx')], timings)
            timings[label_f]['Load Program Binary'] = addTime([(label_ptx, 'Setup')], timings)
            timings[label_f]['Copy Pointer Arguments'] = addTime([(label_ptx, 'Copy_Ptr')], timings)
            timings[label_f]['Copy Scalar Arguments'] = addTime([(label_ptx, 'Copy_Scalar')], timings)
            timings[label_f]['WorkGroup Size Calculation'] = addTime([(label_ptx, 'Misc')], timings)
            timings[label_f]['Kernel'] = addTime([(label_ptx, 'Kernel')], timings)
            timings[label_f]['Read Output'] = addTime([(label_ptx, 'Read_Output'), (label_pb, 'Copy')], timings)
            timings[label_f]['Free Memory'] = addTime([(label_ptx, 'Mem_Free'), (label_pb, 'Mem_Untrack')], timings)
            timings[label_f]['Clear Context'] = addTime([(label_ptx, 'Clear_Ctx')], timings)
            timings[label_f]['Timer Wall Time'] = addTime([(label_pb, 'Timer Wall Time')], timings)
        else:
            #if len(timings) == 1:
            timings[label_f] = copy.deepcopy(timings[label_pb])
        timings[label_f]['Timer Wall - IO'] = timings[label_f]['Timer Wall Time'] - timings[label_f]['IO']

    def parseLogForTimer():
        timings = {}
        log = StringIO.StringIO(globals.outlog)
        for line in log.readlines():
            if line.startswith('Printing Parboil Timer'):
                timerName = 'Parboil'
                timings[timerName] = {}
                continue
            if line.startswith('Printing VISC Timer'):
                regex = re.search('Printing VISC Timer: *(?P<name>[a-zA-Z0-9 _]+)', line)
                timerName = regex.group('name').strip()
                timings[timerName] = {}
                if timerName != 'NVPTX_Timer' and timerName != 'X86_Timer' and timerName != 'GenVISC_Timer' and timerName != 'KernelTimer' and timerName != 'SPIR_Timer':
                    print "Warning: Found unknown timer " + timerName
                continue
            m = re.search('(?P<timerID>[a-zA-Z _/]+) *: *(?P<value>[0-9]*\.[0-9]*) *$', line)
            if m:
                if timings[timerName] is not None:
                    timings[timerName][m.group('timerID').strip()] = float(m.group('value').strip())
                    #print timings[timerName][m.group('timerID').strip()]
                    #print m.group('value').strip()
                    #print
                else:
                    print "Error: Found unknown timer:" + timerName
        if timings is not {}:
            compileTimings(timings)
        return timings

    def writeToCsv(table, filename):
        writer = csv.writer(open(filename, 'wb'))
        writer.writerow(["Application", "Version", "Test", "Timer", "Category", "Time in seconds"]);
        for app, app_record in table.items():
            writer.writerow([])
            for ver, ver_record in app_record.items():
                for test, test_record in ver_record.items():
                    for timer, timer_record in test_record.items():
                        for category, value in timer_record.items():
                            writer.writerow([app, ver, test, timer, category, value]);
                            app = ""
                            ver = ""
                            test = ""
                            timer = ""
                            category = ""
                            value = ""


    def addToTotal(totalTimings, instanceTimings):
        for timer, timer_record in totalTimings.items():
            for category, value in timer_record.items():
                totalTimings[timer][category] += instanceTimings[timer][category]
        return totalTimings

    def averageTime(totalTimings, num_instances):
        for timer, timer_record in totalTimings.items():
            for category, value in timer_record.items():
                totalTimings[timer][category] /= num_instances
        return totalTimings

    def run():
        (opts, pos) = parser.parse_args(args)
        globals.verbose = opts.verbose

        configs = [
                    ('spmv',    {  'VERSION'   : ["opencl_nvidia", "visc"],
                                    'TEST'      : [("large", 10)]
                                }
                    )
                    ,('sgemm',   {  'VERSION'   : ["opencl_nvidia", "visc_sh"],
                                    'TEST'      : [("4K", 10)]
                                }
                    )
                    ,('lbm',     {   'VERSION'   : ["opencl_nvidia", "visc"],
                                    'TEST'      : [("long", 10)]
                                }
                    )
                    ,('stencil', {   'VERSION'   : ["opencl_base", "visc"],
                                    'TEST'      : [("large", 10)]
                                 }
                    )
                    ,('bfs',    {   'VERSION'   : ["opencl_nvidia", "visc"],
                                    'TEST'      : [("1M", 10), ("SF", 10)]
                                }
                    )
                    ,('tpacf',  {   'VERSION'   : ["opencl_base", "visc"],
                                    'TEST'      : [("large", 10)]
                                }
                    )
                    ,('cutcp',  {   'VERSION'   : ["opencl_nvidia", "visc"],
                                    'TEST'      : [("large", 10)]
                                }
                    )
                    #('histo',  {   'VERSION'   : ["opencl_nvidia", "visc"],
                                    #'TEST'      : [("default", 10), ("large", 10)]
                                #}
                    #)
                    #('spmv',    {  'VERSION'   : ["opencl_cpu_baseline", "visc"],
                                    #'TEST'      : [("large", 10), ("huge", 10)]
                                #}
                    #)
                    #('sgemm',   {   'VERSION'   : ["opencl_cpu_sm", "visc_sh"],
                                    #'TEST'       : [("medium", 1), ("4K", 1)]
                                #}
                    #)
                   #('lbm',     {   'VERSION'   : ["opencl_cpu_baseline", "visc"],
                                    #'TEST'      : [("short", 10), ("long", 10)]
                                #}
                    #)
                   #,('stencil', {   'VERSION'   : ["opencl_cpu_baseline", "visc"],
                                    #'TEST'      : [("default", 10), ("large", 10)]
                                 #}
                    #)
                    #('bfs',    {   'VERSION'   : ["opencl_cpu_baseline", "visc_base"],
                                    #'TEST'      : [("1M", 5), ("SF", 5)]
                                #}
                    #)
                    #,('tpacf',  {   'VERSION'   : ["opencl_cpu_base", "visc"],
                                    #'TEST'      : [("medium", 1), ("large", 1)]
                                #}
                    #)
                    #,('cutcp',  {   'VERSION'   : ["opencl_cpu_baseline", "visc"],
                                    #'TEST'      : [("small", 1), ("large", 1)]
                                #}
                    #)
                    #,('histo',  {   'VERSION'   : ["opencl_cpu_baseline", "visc"],
                                    #'TEST'      : [("default", 1), ("large", 1)]
                                #}
                    #)
                  ]
        #log_dir = tempfile.mkdtemp(prefix = 'LOG-', dir = ".")
        log_dir = getLogDirname()
        os.makedirs(log_dir)
        table = {}
        print "Logs stored at: " + log_dir
        for (app, app_config) in configs:
            table[app] = {}
            print
            print "=========== APPLICATION CONFIGURATION ==========="
            print "App Name:\t" + app
            print "Version(s):\t" + "\t".join(app_config['VERSION'])
            print "Test(s):\t" + "\t".join(x[0] for x in app_config['TEST'])
            print "================================================="

            app_dir = log_dir+"/"+app
            os.makedirs(app_dir)

            arguments = [app]
            for ver in app_config['VERSION']:
                table[app][ver] = {}
                ver_dir = app_dir+"/"+ver
                os.makedirs(ver_dir)

                arguments.append(ver)
                if globals.verbose:
                    arguments.append("-v")

                print "Cleaning " + app + " " + ver
                action = clean_options(progname, 'clean', arguments).run()
                #action = False
                if action:
                    action()
                else:
                    print "Could not clean: " + app

                dump_cleanLog(ver_dir)

                #arguments.append("TIMER=no")
                print "Compiling " + app + " " + ver + " with arguments " + " ".join(arguments[2:])
                action = compile_options(progname, 'compile', arguments).run()
                #action = False
                if action:
                    action()
                else:
                    print "Could not compile: " + app
                dump_buildLog(ver_dir)

                if globals.verbose:
                    arguments.pop()

                for test, num_instances in app_config['TEST']:
                    print ("Running " + app + "-" + ver + " with test " + test + " "),
                    arguments.append(test)
                    if globals.verbose:
                        arguments.append("-v")
                        arguments.append("-C")
                    for i in range(0, num_instances):
                        print('.'),
                        sys.stdout.flush()
                        action = run_options(progname, 'run', arguments).run()
                        if action:
                            action()
                        else:
                            print "Could not run: " + app

                        instanceTimings = parseLogForTimer()
                        if i == 0:
                            table[app][ver][test] = instanceTimings
                        else:
                            table[app][ver][test] = addToTotal(table[app][ver][test], instanceTimings)

                    table[app][ver][test] = averageTime(table[app][ver][test], num_instances)
                    dump_runLog(ver_dir, test)
                    if globals.verbose:
                        arguments.pop()
                        arguments.pop()
                    arguments.pop()
                    print

                arguments.pop()
        writeToCsv(table, log_dir + "/table.csv")
        return None

    return OptionGetter(parser.print_help, run)

# Dictionary from option name to function from command-line parameters
# to pair of help thunk and option processor thunk
parse_mode_options = {
    'help'     : help_options,
    '-h'       : help_options,
    '--help'   : help_options,
    'list'     : list_options,
    'describe' : describe_options,
    'clean'    : clean_options,
    'compile'  : compile_options,
    'run'      : run_options,
    'debug'    : debug_options,
    'time'     : time_options
    }

def parse_options(args):
    """Parse a list of command-line options.  If there is an error in
    the options, then the function will print a message and either call
    sys.exit() or return None.  If options were parsed successfully
    then it will return a thunk which represents the action to take,
    or None if no action need be taken.

    Generally, the caller should call the return value, unless None is
    returned."""
    # Parse arguments; skip the program name
    prog_name = args[0]

    # Get the command name
    try: cmd = args[1]
    except IndexError: cmd = 'help'

    # Dispatch
    try: mode = parse_mode_options[cmd]
    except KeyError:
        invalid_option_message(prog_name, cmd, args[2:])
        return

    # Set up and run the option parser
    return mode(prog_name, cmd, args[2:]).run()


